#include <stdlib.h>

#include "output.h"
#include "optimize.h"
#include "buffer.h"
#include "state.h"

/* reorders previous tags based on the end tags produced at the
   current location, returns the new number of elements that must be
   popped off */
unsigned int
reorder (Block *block, int prev_pop, Buffer *pop_stack, Buffer *tag_stack)
{
    unsigned int t, n_pop;
    Buffer *stack;
    Buffer *fixed;
    Buffer *common;

    if (block == NULL || empty (pop_stack))
	return 0;

    stack = buffer_dup (block->stack);
    fixed = new_buffer (8);
    t = prev_pop;
    while (t-- && !empty(stack)) {
	push_char (fixed, pop_char (stack));
    }
    common = intersect (stack, pop_stack);

    if (empty (common)) {
	add (tag_stack, stack);
	n_pop = stack->index + 1;
    } else if (common->index == stack->index) {
	subtract (pop_stack, common);
	n_pop = stack->index + 1;
    } else if (common->index < pop_stack->index) {
	subtract (pop_stack, common);
	subtract (stack, common);
	add (tag_stack, stack);
	add (stack, common);
	n_pop = stack->index + 1;
    } else {
	subtract (stack, common);
	add (stack, common);
	clear_buffer (pop_stack);
	n_pop = common->index + 1;
    }

    while (prev_pop && !empty(fixed)) {
	push_char (stack, pop_char (fixed));
	prev_pop--;
    }

    free_buffer (block->stack);
    block->stack = stack;
    free_buffer (common);

    return n_pop + reorder (block->prev, prev_pop + block->n_pop, pop_stack, tag_stack);
}


/* simplified version of reorder */
void
order (Block *block, int prev_pop, Buffer *pop_stack)
{
    unsigned int t, n_pop;
    Buffer *stack;
    Buffer *fixed;
    Buffer *common;

    if (block == NULL || empty (pop_stack))
	return;

    stack = buffer_dup (block->stack);
    fixed = new_buffer (8);
    t = prev_pop;
    while (t-- && !empty(stack)) {
	push_char (fixed, pop_char (stack));
    }
    common = intersect (stack, pop_stack);

    if (empty (common)) {
    } else if (common->index == stack->index) {
	subtract (pop_stack, common);
	n_pop = stack->index + 1;
    } else if (common->index < pop_stack->index) {
	subtract (pop_stack, common);
	subtract (stack, common);
	add (stack, common);
    } else {
	subtract (stack, common);
	add (stack, common);
	clear_buffer (pop_stack);
    }

    while (prev_pop && !empty(fixed)) {
	push_char (stack, pop_char (fixed));
	prev_pop--;
    }

    free_buffer (block->stack);
    block->stack = stack;
    free_buffer (common);

    return order (block->prev, prev_pop + block->n_pop, pop_stack);
}

unsigned int
end_tag_sizes (Block *block, unsigned int n)
{
    unsigned int length = 0;
    Buffer *temp;
    int c;

    if (n == 0 || block == NULL)
	return 0;

    temp = new_buffer (8);
 
    for (;n != 0 && !empty(block->stack); n--) {
	switch (c = pop_char (block->stack)) {
	case 0: length += 4; break;
	case 1: length += 4; break;
	case 2: length += 4; break;
	case 3: length += 4; break;
	case 4: length += 4; break;
	case 5: length += 4; break;
	case 6: length += 4; break;
	case 7: length += 4; break;
	case 8: length += 4; break;
	case 9: length += 4; break;
	case B: length += 4; break;
	case EM: length += 5; break;
	case I: length += 4; break;
	case PL: length += 5; break;
	case S: length += 4; break;
	case TT: length += 5; break;
	case U: length += 4; break;
	case cr: length += 4; break;
	case cg: length += 4; break;
	case cb: length += 4; break;
	case cc: length += 4; break;
	case cm: length += 4; break;
	case cy: length += 4; break;
	case ck: length += 4; break;
	case cw: length += 4; break;	
	}
	push_char (temp, c);
    }

    while (!empty (temp)) {
	push_char (block->stack, pop_char (temp));
    }

    free_buffer (temp);

    return length + end_tag_sizes (block->prev, n);
}

unsigned int
get_block_size (Block *block)
{
    int i;
    unsigned int length = 0;

    for (i = 0; i <= block->stack->index; i++) {
	switch (index (block->stack, i)) {
	case 0: length += 3; break;
	case 1: length += 3; break;
	case 2: length += 3; break;
	case 3: length += 3; break;
	case 4: length += 3; break;
	case 5: length += 3; break;
	case 6: length += 3; break;
	case 7: length += 3; break;
	case 8: length += 3; break;
	case 9: length += 3; break;
	case B: length += 3; break;
	case EM: length += 4; break;
	case I: length += 3; break;
	case PL: length += 4; break;
	case S: length += 3; break;
	case TT: length += 4; break;
	case U: length += 3; break;
	case cr: length += 3; break;
	case cg: length += 3; break;
	case cb: length += 3; break;
	case cc: length += 3; break;
	case cm: length += 3; break;
	case cy: length += 3; break;
	case ck: length += 3; break;
	case cw: length += 3; break;	
	}
    }

    return length + end_tag_sizes (block->prev, block->n_pop);
}

typedef struct _Combination {
    struct _Combination *next;
    Block *block;
    Buffer *size_stack;
    Buffer *color_stack;
    Buffer *pop_stack;
    unsigned int n_pl;
} Combination;

Combination *
push_combination (Combination *c, char *text, Buffer *size_stack,
		  Buffer *color_stack, unsigned int n_pl)
{
    Combination *n = malloc (sizeof (Combination));

    n->block = malloc (sizeof (Block));
    n->block->next = n->block->prev = NULL;
    n->block->text = text;
    n->next = c;
    n->size_stack = buffer_dup (size_stack);
    n->color_stack = buffer_dup (color_stack);

    return n;
}

void
free_combinations (Combination *c)
{
    Combination *next;

    if (c == NULL)
	return;
    
    next = c->next;
    free_buffer (c->size_stack);
    free_buffer (c->color_stack);
    free_buffer (c->pop_stack);
    if (c->block) {
	free_buffer (c->block->stack);
	free (c->block);
    }
    free (c);

    free_combinations (next);
}

Combination *
get_combinations (TaggedText *text, BlockList *block_list,
		  Buffer *size_stack, Buffer *color_stack, int n_pl)
{
    State previous_state = INITIAL_STATE, state;
    Buffer *pop_stack = new_buffer (8);
    Buffer *tag_stack = new_buffer (8);
    Combination *comb, *combinations;
    unsigned int i;

    comb = combinations = push_combination (NULL, text->text, size_stack, color_stack, n_pl);

    if (text->prev)
	previous_state = text->prev->state;

    state = text->state;
    
    if (previous_state.b == 0 && state.b == 1)
	push_char (tag_stack, B);
    else if (previous_state.b == 1 && state.b == 0)
	push_char (pop_stack, B);
  
    if (previous_state.s == 0 && state.s == 1)
	push_char (tag_stack, S);
    else if (!(previous_state.s == 1 && state.s == 1)) {
	if (previous_state.s == 1 && state.s == 0)
	    push_char (pop_stack, S);
	if (previous_state.em == 0 && state.em == 1)
	    push_char (tag_stack, EM);
	else if (previous_state.em == 1 && state.em == 0)
	    push_char (pop_stack, EM);
    }

    if (previous_state.i == 0 && state.i == 1)
	push_char (tag_stack, I);
    else if (previous_state.i == 1 && state.i == 0)
	push_char (pop_stack, I);

    if (previous_state.tt == 0 && state.tt == 1)
	push_char (tag_stack, TT);
    else if (previous_state.tt == 1 && state.tt == 0)
	push_char (pop_stack, TT);

    if (previous_state.u < state.u) {
	for (i = 0; i < state.u - previous_state.u; i++)
	    push_char (tag_stack, U);
    } else if (previous_state.u > state.u) {
	for (i = 0; i < previous_state.u - state.u; i++)
	    push_char (pop_stack, U);
    }

    if (previous_state.size != state.size) {
	if (size_stack->index > 0
	    && next_char (size_stack) == state.size) {
	    push_char (pop_stack, pop_char (comb->size_stack));
	} else {
	    push_char (tag_stack, state.size);
	    push_char (comb->size_stack, state.size);
	}
    }

    if (previous_state.color != state.color) {
	if (color_stack->index > 0
	    && next_char (color_stack) == 17+state.color) {
	    push_char (pop_stack, pop_char (comb->color_stack));
	} else {
	    push_char (tag_stack, 17 + state.color);
	    push_char (comb->color_stack, 17 + state.color);
	}
    }

    comb->pop_stack = buffer_dup (pop_stack);
    comb->block->n_pop = reorder (block_list->tail, 0, pop_stack, tag_stack);
    free_buffer (pop_stack);
    comb->block->stack = tag_stack;

    return combinations;
}

/* returns a block with the minimum */
unsigned int
test_combination (TaggedText *text, BlockList *block_list,
		  Buffer *size_stack, Buffer *color_stack,
		  unsigned int n_pl, int look_ahead)
{
    Combination *combinations, *comb;
    Block *block;
    Buffer *pop_stack = new_buffer (8);
    unsigned int block_size, size, min = 0xffffffff;

    if (text == NULL)
	return end_tag_sizes (block_list->tail, 0xffffffff);
    else if (look_ahead == 0)
	return 0;

    comb = combinations = get_combinations (text, block_list, size_stack, color_stack, n_pl);

    do {
	block = comb->block;
	clear_buffer (pop_stack);
	add (pop_stack, comb->pop_stack);
	order (block_list->tail, 0, pop_stack);
	append_block (block_list, block);
	block_size = get_block_size (block);
	size = test_combination (text->next, block_list,
				 comb->size_stack, comb->color_stack, n_pl, look_ahead-1);
	if (block_size + size < min)
	    min = block_size + size;
	detach_block (block_list);
    } while (comb = comb->next);

    free_combinations (combinations);
    free_buffer (pop_stack);

    return min;
}

unsigned int
is_stack_clear (Block *block, int n)
{
    
    if (block == NULL) {
	if (n == 0)
	    return 1;
	else
	    return 0;
    }

    n -= block->stack->index+1;

    if (n < 0)
	return 0;

    return is_stack_clear (block->prev, n + block->n_pop);
}

/* Look ahead is an area to test the affect of different operator
   combinations. This specifies the number of blocks to look ahead. */
#define LOOK_AHEAD 5

/* Simple optimizer. It trys to find the minimum number of tags to
   output to change the state appropriately for the next piece of
   text. */
Block *
optimize (TaggedTextList *text_list)
{
    TaggedText *cur = text_list->head;
    Buffer *size_stack = new_buffer (8);
    Buffer *color_stack = new_buffer (8);
    Buffer *pop_stack = new_buffer (8);
    BlockList block_list = {0};
    Combination *combinations, *comb, *smallest_comb;
    Block *block, *next;
    unsigned int block_size, size = 0, min, n_pl;

    push_char (size_stack, DEFAULT_SIZE); 
    push_char (color_stack, 17+DEFAULT_COLOR);

    for (; cur != NULL; cur = cur->next) {
	min = 0xffffffff;

	comb = combinations = get_combinations (cur, &block_list, size_stack, color_stack, n_pl);
	
	do {
	    block = comb->block;
	    clear_buffer (pop_stack);
	    add (pop_stack, comb->pop_stack);
	    order (block_list.tail, 0, pop_stack);
	    append_block (&block_list, block);
	    block_size = get_block_size (block);
	    /*size = test_combination (cur->next, &block_list, size_stack, color_stack, n_pl, LOOK_AHEAD);*/
	    if (block_size + size < min) {
		min = block_size + size;
		smallest_comb = comb;
	    }
	    detach_block (&block_list);
	} while (comb = comb->next);

	clear_buffer (size_stack);
	add (size_stack, smallest_comb->size_stack);
	clear_buffer (color_stack);
	add (color_stack, smallest_comb->color_stack);
	clear_buffer (pop_stack);
	add (pop_stack, smallest_comb->pop_stack);
	order (block_list.tail, 0, pop_stack);
	n_pl = smallest_comb->n_pl;
	append_block (&block_list, smallest_comb->block);
	smallest_comb->block = NULL;
	free_combinations (combinations);

	/* intermediate output */
	if (block_list.tail->prev != NULL
	    && is_stack_clear (block_list.tail->prev, block_list.tail->n_pop)) {
	    block_list.tail->prev->next = NULL;
	    block_list.tail->prev = NULL;
	    block_list.tail->n_pop = 0;
	    output (block_list.head);
	    block_list.head = block_list.tail;
	    cur->prev->next = NULL;
	    cur->prev = NULL;
	    free_text_list (text_list->head);
	    text_list->head = cur;
	}
    }

    free_buffer (size_stack);
    free_buffer (color_stack);
    free_buffer (pop_stack);

    return block_list.head;
}
